@@ -3,36 +3,27 @@ import enum
 import textwrap
 import inspect
 from functools import wraps
+import functools
+import operator
 from collections import namedtuple
 import os
 
 import six
-from .config import get_compile_dir, set_compile_dir
-# TODO: Remove circular dependency required for `circuit.bind` logic
-import magma as m
 from . import cache_definition
-from .clock import ClockTypes
-from .common import deprecated, Stack
+from .common import deprecated, setattrs, Stack, OrderedIdentitySet
 from .interface import *
 from .wire import *
 from .config import get_debug_mode
 from .debug import get_callee_frame_info, debug_info
 from .logging import root_logger
 from .is_definition import isdefinition
-from .ref import AnonRef, ArrayRef, TupleRef, DefnRef, InstRef
-from .bit import VCC, GND
-from .array import Array
 from .placer import Placer, StagedPlacer
-from .tuple import Tuple
-from .digital import Digital
 from magma.syntax.combinational import combinational
 from magma.syntax.sequential import sequential
 from magma.syntax.verilog import combinational_to_verilog, \
     sequential_to_verilog
-from .verilog_utils import (value_to_verilog_name,
-                            convert_values_to_verilog_str,
-                            process_inline_verilog)
 from .view import PortView
+from magma.protocol_type import MagmaProtocol
 
 __all__ = ['AnonymousCircuitType']
 __all__ += ['AnonymousCircuit']
@@ -41,15 +32,31 @@ __all__ += ['CircuitType']
 __all__ += ['Circuit']
 __all__ += ['DeclareCircuit']
 __all__ += ['DefineCircuit', 'EndDefine', 'EndCircuit']
+__all__ += ['DefineCircuitKind']
 
 __all__ += ['CopyInstance']
 __all__ += ['circuit_type_method']
 __all__ += ['circuit_generator']
+__all__ += ['CircuitBuilder', 'builder_method']
 
 circuit_type_method = namedtuple('circuit_type_method', ['name', 'definition'])
 
 _logger = root_logger()
 
+_VERILOG_FILE_OPEN = """
+integer \_file_{filename} ;
+initial \_file_{filename} = $fopen(\"{filename}\", \"{mode}\");
+"""  # noqa
+
+_VERILOG_FILE_CLOSE = """
+final $fclose(\_file_{filename} );
+"""  # noqa
+
+_DEFAULT_VERILOG_LOG_STR = f"""
+`ifndef MAGMA_LOG_LEVEL
+    `define MAGMA_LOG_LEVEL 1
+`endif"""
+
 
 class _SyntaxStyle(enum.Enum):
     NONE = enum.auto()
@@ -61,19 +68,53 @@ class DefinitionContext:
     def __init__(self, placer):
         self.placer = placer
         self._inline_verilog = []
+        self._displays = []
+        self._insert_default_log_level = False
+        self._files = []
+        self._builders = []
+
+    def add_builder(self, builder):
+        self._builders.append(builder)
+
+    def add_file(self, file):
+        self._files.append(file)
+
+    def _finalize_file_opens(self):
+        for file in self._files:
+            self.add_inline_verilog(
+                _VERILOG_FILE_OPEN.format(filename=file.filename,
+                                          mode=file.mode),
+                {}, {})
+
+    def _finalize_file_close(self):
+        for file in self._files:
+            self.add_inline_verilog(
+                _VERILOG_FILE_CLOSE.format(filename=file.filename),
+                {}, {})
 
     def add_inline_verilog(self, format_str, format_args, symbol_table):
         self._inline_verilog.append((format_str, format_args, symbol_table))
 
+    def insert_default_log_level(self):
+        self._insert_default_log_level = True
+
+    def add_display(self, display):
+        self._displays.append(display)
+
+    def _finalize_displays(self):
+        for display in self._displays:
+            self.add_inline_verilog(*display.get_inline_verilog())
+
     def finalize(self, defn):
-        final_placer = self.placer.finalize(defn)
-        final_context = DefinitionContext(final_placer)
-        # inline logic may introduce instances
-        with _DefinitionContextManager(final_context):
-            for format_str, format_args, symbol_table in self._inline_verilog:
-                process_inline_verilog(defn, format_str, format_args,
-                                       symbol_table)
-        return final_context
+        self.placer = self.placer.finalize(defn)
+        for builder in self._builders:
+            inst = builder.finalize()
+            self.placer.place(inst)
+        if self._insert_default_log_level:
+            self.add_inline_verilog(_DEFAULT_VERILOG_LOG_STR, {}, {})
+        self._finalize_file_opens()  # so displays can refer to open files
+        self._finalize_displays()
+        self._finalize_file_close()  # close after displays
 
 
 _definition_context_stack = Stack()
@@ -90,28 +131,21 @@ class _DefinitionContextManager:
         _definition_context_stack.pop()
 
 
-def _setattrs(obj, dct):
-    for k, v in dct.items():
-        if isinstance(k, str):
-            setattr(obj, k, v)
-
-
 def _has_definition(cls, port=None):
     if cls.instances:
         return True
+    if cls._context_._inline_verilog:
+        return True
     if port is None:
         interface = getattr(cls, "interface", None)
         if not interface:
-            return None
+            return False
         return any(_has_definition(cls, p) for p in interface.ports.values())
-    if isinstance(port, Tuple) or isinstance(port, Array):
-        return any(_has_definition(cls, elem) for elem in port)
-    if not port.is_output():
-        return port.value() is not None
-    return False
+    flat = port.flatten()
+    return any(not f.is_output() and f.value() is not None for f in flat)
 
 
-def _get_interface_decl(cls):
+def _get_interface_type(cls):
     if hasattr(cls, "IO") and hasattr(cls, "io"):
         _logger.warning("'IO' and 'io' should not both be specified, ignoring "
                         "'io'", debug_info=cls.debug_info)
@@ -124,34 +158,26 @@ def _get_interface_decl(cls):
                         debug_info=cls.debug_info)
         if isinstance(cls.IO, InterfaceKind):
             return None
-        return DeclareInterface(*cls.IO)
+        return make_interface(*cls.IO)
     if hasattr(cls, "io"):
         cls._syntax_style_ = _SyntaxStyle.NEW
-        return DeclareLazyInterface(cls.io)
+        return cls.io.make_interface()
     return None
 
 
-def _add_intermediate_value(value, values):
+def _add_intermediate_value(value):
     """
     Add an intermediate value `value` to `values`, handling members of
     recursive types.  Used by `_get_intermediate_values` as part of
     `CircuitKind.__repr__`.
     """
-    # If we encounter a member of an array or tuple, add the entire parent
-    # value (only once)
-    if isinstance(value.name, ArrayRef):
-        _add_intermediate_value(value.name.array, values)
-    elif isinstance(value.name, TupleRef):
-        _add_intermediate_value(value.name.tuple, values)
-    elif not isinstance(value.name, (DefnRef, InstRef, AnonRef)):
-        if value is VCC or value is GND:
-            # Skip VCC and GND because they are special
-            return
-        if not any(value is x for x in values):
-            values.append(value)
+    root = value.name.root()
+    if root is None or root.const():
+        return OrderedIdentitySet()
+    return OrderedIdentitySet((root,))
 
 
-def _get_intermediate_values(value, values):
+def _get_intermediate_values(value):
     """
     Retrieve the intermediate values connected to `value` and put them into
     `values`
@@ -160,26 +186,28 @@ def _get_intermediate_values(value, values):
     values in a circuit.
     """
     if value.is_output():
-        return
+        return OrderedIdentitySet()
     if value.is_mixed():
         # Mixed
-        for v in value:
-            _get_intermediate_values(v, values)
-        return
+        return functools.reduce(operator.or_,
+                                (_get_intermediate_values(v) for v in value),
+                                OrderedIdentitySet())
     driver = value.value()
     if driver is None:
-        return
-    if isinstance(value, (Array, Tuple)) and driver.name.anon():
-        for elem in value:
-            _get_intermediate_values(elem, values)
-    else:
-        while driver is not None:
-            _add_intermediate_value(driver, values)
-            if not driver.is_output():
-                value = driver
-                driver = driver.value()
-            else:
-                driver = None
+        return OrderedIdentitySet()
+    flat = value.flatten()
+    if len(flat) > 1 and driver.name.anon():
+        return functools.reduce(operator.or_,
+                                (_get_intermediate_values(f) for f in flat),
+                                OrderedIdentitySet())
+    values = OrderedIdentitySet()
+    while driver is not None:
+        values |= _add_intermediate_value(driver)
+        if not driver.is_input():
+            break
+        value = driver
+        driver = driver.value()
+    return values
 
 
 class CircuitKind(type):
@@ -198,7 +226,9 @@ class CircuitKind(type):
         dct.setdefault('primitive', False)
         dct.setdefault('coreir_lib', 'global')
         dct["inline_verilog_strs"] = []
+        dct["inline_verilog_generated"] = False
         dct["bind_modules"] = {}
+        dct["bind_modules_bound"] = False
 
         # If in debug_mode is active and debug_info is not supplied, attach
         # callee stack info.
@@ -217,16 +247,18 @@ class CircuitKind(type):
 
         # Create interface for this circuit class.
         cls._syntax_style_ = _SyntaxStyle.NONE
-        IO = _get_interface_decl(cls)
+        IO = _get_interface_type(cls)
         if IO is not None:
             cls.IO = IO
-            cls.interface = cls.IO(defn=cls,
-                                   renamed_ports=dct["renamed_ports"])
-            _setattrs(cls, cls.interface.ports)
+            cls.interface = cls.IO(defn=cls, renamed_ports=dct["renamed_ports"])
+            setattrs(cls, cls.interface.ports, lambda k, v: isinstance(k, str))
         try:
             context = _definition_context_stack.pop()
             assert context.placer.name == cls_name
-            cls._context_ = context.finalize(cls)
+            # Override staged context with '_context_' from namespace if
+            # available.
+            cls._context_ = dct.get("_context_", context)
+            cls._context_.finalize(cls)
         except IndexError:  # no staged placer
             cls._context_ = DefinitionContext(Placer(cls))
 
@@ -266,14 +298,15 @@ class CircuitKind(type):
         for instance in sorted_instances:
             values.extend(list(instance.interface.ports.values()))
         values.extend(cls.interface.ports.values())
-        intermediate_values = []
-        for value in values:
-            _get_intermediate_values(value, intermediate_values)
+        intermediate_values = (_get_intermediate_values(value)
+                               for value in values)
+        intermediate_values = functools.reduce(operator.or_,
+                                               intermediate_values,
+                                               OrderedIdentitySet())
         if intermediate_values:
-            s += "\n".join(
-                f"{value.name} = {repr(value)}" for value in intermediate_values
-            ) + "\n"
-
+            intermediate_values = (f"{value.name} = {repr(value)}"
+                                   for value in intermediate_values)
+            s += "\n".join(intermediate_values) + "\n"
         # Emit instances.
         for instance in sorted_instances:
             s += repr(instance) + '\n'
@@ -320,16 +353,10 @@ class CircuitKind(type):
             defn[name] = cls
         return defn
 
-    def _inline_verilog(cls, inline_str, **kwargs):
-        format_args = {}
-        for key, arg in kwargs.items():
-            format_args[key] = convert_values_to_verilog_str(arg)
-        cls.inline_verilog_strs.append(inline_str.format(**format_args))
-
     @deprecated(msg="cls.inline_verilog is deprecated, use m.inline_verilog"
                 "instead")
     def inline_verilog(cls, inline_str, **kwargs):
-        cls._inline_verilog(inline_str, **kwargs)
+        cls._context_.add_inline_verilog(inline_str, kwargs, None)
 
 
 @six.add_metaclass(CircuitKind)
@@ -401,12 +428,12 @@ class AnonymousCircuitType(object):
             # Wire the circuit's outputs to this circuit's inputs.
             self.wireoutputs(output.interface.outputs(), debug_info)
             return
-        if isinstance(output, m.MagmaProtocol):
+        if isinstance(output, MagmaProtocol):
             output = output._get_magma_value_()
         # Wire the output to this circuit's input (should only have 1 input).
         inputs = []
         for inp in self.interface.inputs():
-            if isinstance(inp, m.MagmaProtocol):
+            if isinstance(inp, MagmaProtocol):
                 inp = inp._get_magma_value_()
             inputs.append(inp)
         ni = len(inputs)
@@ -461,7 +488,7 @@ class AnonymousCircuitType(object):
         return o[0] if len(o) == 1 else tuple(o)
 
     def setinterface(self, interface):
-        _setattrs(self, interface.ports)
+        setattrs(self, interface.ports, lambda k, v: isinstance(k, str))
         self.interface = interface
         return self
 
@@ -574,9 +601,10 @@ class DefineCircuitKind(CircuitKind):
             for base in bases:
                 if base is Circuit:
                     continue
-                if not issubclass(base, Circuit):
-                    raise Exception(f"Must subclass from Circuit or a "
-                                    f"subclass of Circuit ({base})")
+                if not issubclass(base, AnonymousCircuitType):
+                    raise Exception(f"Must subclass from AnonymousCircuitType "
+                                    f"or a subclass of AnonymousCircuitType "
+                                    f"({base})")
                 dct['name'] = base.name
                 break
         name = dct['name']
@@ -646,49 +674,8 @@ class DefineCircuitKind(CircuitKind):
         """Place a circuit instance in this definition"""
         cls._context_.placer.place(inst)
 
-    def gen_bind_port(cls, mon_arg, bind_arg):
-        if isinstance(mon_arg, Tuple) or isinstance(mon_arg, Array) and \
-                not issubclass(mon_arg.T, m.Digital):
-            result = []
-            for child1, child2 in zip(mon_arg, bind_arg):
-                result += cls.gen_bind_port(child1, child2)
-            return result
-        port = value_to_verilog_name(mon_arg)
-        arg = value_to_verilog_name(bind_arg)
-        return [(f".{port}({arg})")]
-
     def bind(cls, monitor, *args):
-        bind_str = monitor.verilogFile
-
-        ports = []
-        for mon_arg, cls_arg in zip(monitor.interface.ports.values(),
-                                    cls.interface.ports.values()):
-            if str(mon_arg.name) != str(cls_arg.name):
-                error_str = f"""
-Bind monitor interface does not match circuit interface
-    Monitor Ports: {list(monitor.interface.ports)}
-    Circuit Ports: {list(cls.interface.ports)}
-"""
-                raise TypeError(error_str)
-            ports += cls.gen_bind_port(mon_arg, cls_arg)
-        extra_mon_args = list(
-            monitor.interface.ports.values()
-        )[len(cls.interface):]
-        for mon_arg, bind_arg in zip(extra_mon_args, args):
-            ports += cls.gen_bind_port(mon_arg, bind_arg)
-        ports_str = ",\n    ".join(ports)
-        bind_str = f"bind {cls.name} {monitor.name} {monitor.name}_inst (\n    {ports_str}\n);"  # noqa
-        if not os.path.isdir(".magma"):
-            os.mkdir(".magma")
-        curr_compile_dir = get_compile_dir()
-        set_compile_dir("normal")
-        # Circular dependency, need coreir backend to compile, backend imports
-        # circuit (for wrap casts logic, we might be able to factor that out)
-        m.compile(f".magma/{monitor.name}", monitor, inline=True)
-        set_compile_dir(curr_compile_dir)
-        with open(f".magma/{monitor.name}.v", "r") as f:
-            content = "\n".join((f.read(), bind_str))
-        cls.bind_modules[monitor.name] = content
+        cls.bind_modules[monitor] = args
 
 
 @six.add_metaclass(DefineCircuitKind)
@@ -747,16 +734,18 @@ def CopyInstance(instance):
     return new_instance
 
 
-GeneratorArguments = namedtuple('GeneratorArguments', ['args', 'kwargs'])
+_GeneratorArguments = namedtuple('GeneratorArguments', ['args', 'kwargs'])
 
 
+@deprecated(msg="circuit_generator decorator is deprecated, subclass Generator "
+            "instead")
 def circuit_generator(func):
     @cache_definition
     @wraps(func)
     def wrapped(*args, **kwargs):
         result = func(*args, **kwargs)
         # Store arguments to generate the circuit.
-        result._generator_arguments = GeneratorArguments(args, kwargs)
+        result._generator_arguments = _GeneratorArguments(args, kwargs)
         return result
     return wrapped
 
@@ -774,3 +763,62 @@ coreir_port_mapping = {
 def DeclareCoreirCircuit(*args, **kwargs):
     return DeclareCircuit(*args, **kwargs,
                           renamed_ports=coreir_port_mapping)
+
+
+class _CircuitBuilderMeta(type):
+    pass
+
+
+def builder_method(func):
+
+    @wraps(func)
+    def _wrapped(this, *args, **kwargs):
+        with _DefinitionContextManager(this._context):
+            result = func(this, *args, **kwargs)
+        return result
+
+    return _wrapped
+
+
+class CircuitBuilder(metaclass=_CircuitBuilderMeta):
+    _RESERVED_NAMESPACE_KEYS = {"io", "_context_", "name"}
+
+    def __init__(self, name):
+        try:
+            context = _definition_context_stack.peek()
+            context.add_builder(self)
+        except IndexError:
+            raise Exception("Can not instance a circuit builder outside a "
+                            "definition")
+        self._name = name
+        self._io = SingletonInstanceIO()
+        self._finalized = False
+        self._dct = {}
+        self._context = DefinitionContext(StagedPlacer(self._name))
+
+    def _port(self, name):
+        return self._io.ports[name]
+
+    def _add_port(self, name, typ):
+        self._io.add(name, typ)
+        setattr(self, name, self._io.inst_ports[name])
+
+    def _set_namespace_key(self, key, value):
+        if key in CircuitBuilder._RESERVED_NAMESPACE_KEYS:
+            raise Exception(f"Can not set reserved namespace key '{key}'")
+        self._dct[key] = value
+
+    def _finalize(self):
+        pass
+
+    def finalize(self):
+        if self._finalized:
+            raise Exception("Can only call finalize on a CircuitBuilder once")
+        self._finalize()
+        bases = (AnonymousCircuitType,)
+        dct = {"io": self._io, "_context_": self._context, "name": self._name}
+        dct.update(self._dct)
+        DefineCircuitKind.__prepare__(self._name, bases)
+        t = DefineCircuitKind(self._name, bases, dct)
+        self._finalized = True
+        return t(name=self._name)
