@@ -1,7 +1,4 @@
-import magma as m
-from .bits import BitsMeta, Bits
-from enum import Enum as pyEnum
-
+from magma.bits import BitsMeta, Bits
 
 
 class EnumMeta(BitsMeta):
@@ -27,18 +24,9 @@ class EnumMeta(BitsMeta):
         type_ = cls[num_bits]
         type_._is_magma_enum = True
         for key, value in fields.items():
-            setattr(type_, key, m.bits(value, num_bits))
+            setattr(type_, key, Bits[num_bits](value))
         return type_
 
 
 class Enum(Bits, metaclass=EnumMeta):
     pass
-
-
-# def Enum(**kwargs):
-#     max_value = max(value for value in kwargs.values())
-#     num_bits = max(max_value.bit_length(), 1)
-#     type_ = m.Bits[num_bits]
-#     for key, value in kwargs.items():
-#         setattr(type_, key, m.bits(value, num_bits))
-#     return type_
