@@ -2,6 +2,7 @@ import inspect
 import typing
 import ast
 from .util import get_ast
+from ..protocol_type import MagmaProtocol
 import astor
 import magma.ast_utils as ast_utils
 from magma.debug import debug_info
@@ -35,6 +36,8 @@ class RewriteSelfAttributes(ast.NodeTransformer):
             ret = []
             ports = self.initial_value_map[attr][3].interface.inputs()
             for name, value in zip(ports, node.args):
+                if isinstance(name, MagmaProtocol):
+                    name = name._get_magma_value_()
                 ret.append(ast.parse(
                     f"self_{attr}_{name} = {astor.to_source(value).rstrip()}"
                 ).body[0])
@@ -42,7 +45,10 @@ class RewriteSelfAttributes(ast.NodeTransformer):
             func = astor.to_source(node.func).rstrip()
             outputs = self.initial_value_map[attr][3].interface.outputs()
             if len(outputs) == 1:
-                return ast.Name(f"self_{attr}_{outputs[0]}", ast.Load())
+                name = outputs[0]
+                if isinstance(name, MagmaProtocol):
+                    name = name._get_magma_value_()
+                return ast.Name(f"self_{attr}_{name}", ast.Load())
             else:
                 assert outputs, "Expected module with at least one output"
                 return ast.Tuple([ast.Name(f"self_{attr}_{output}",
@@ -98,6 +104,8 @@ class RewriteReturn(ast.NodeTransformer):
                 for port in eval_value.interface.inputs():
                     if isinstance(port, (m.Clock, m.AsyncReset)):
                         continue
+                    if isinstance(port, MagmaProtocol):
+                        port = port._get_magma_value_()
                     elts.append(ast.Name(f"self_{name}_{port}", ast.Load()))
         if isinstance(node.value, ast.Tuple):
             elts.extend(node.value.elts)
@@ -153,35 +161,6 @@ def get_initial_value_map(init_func, defn_env):
     return initial_value_map
 
 
-def get_io(call_def):
-    """
-    Parses a __call__ method of the form
-
-        def __call__(self, I):
-            O = self.y
-            self.y = self.x
-            self.x = I
-            return O
-
-    Returns a tuple
-    [0]: the list of tuples containing the name and type of each input argument
-         excluding `self`
-    [1]: the output type
-    """
-    # Only support basic args for now
-    assert not call_def.args.vararg
-    assert not call_def.args.kwonlyargs
-    assert not call_def.args.kwarg
-    # assert not call_def.args.defaults
-    assert not call_def.args.kw_defaults
-
-    # Skips self
-    assert call_def.args.args[0].arg == "self"
-    inputs = [(arg.arg, arg.annotation) for arg in call_def.args.args[1:]]
-    returns = call_def.returns
-    return inputs, call_def.returns
-
-
 circuit_definition_template = """
 from magma import Bit, Array, Tuple, Product, Bits, SInt, UInt, IO
 
@@ -261,22 +240,78 @@ def gen_register_instances(initial_value_map, async_reset, magma_name):
     return register_instances
 
 
-def gen_io_args(inputs, output_type, async_reset, magma_name):
+class TypeNamer:
+    def __init__(self, tree, env):
+        self.tree = tree
+        self.env = env
+        self.cache = {}
+
+    def __call__(self, t):
+        try:
+            return self.cache[t]
+        except KeyError:
+            pass
+
+        # Try to keep the name the same
+        for k, v in self.env.items():
+            if v is t:
+                return self.cache.setdefault(t, k)
+
+        # Maybe we can exec the repr?
+        try:
+            s = eval(repr(t), env)
+        except:
+            pass
+        else:
+            if s is t:
+                return self.cache.setdefault(t, repr(t))
+
+        # Or the str?
+        try:
+            s = eval(str(t), env)
+        except:
+            pass
+        else:
+            if s is t:
+                return self.cache.setdefault(t, str(t))
+
+        # give up generate a fesh one
+        n = self.cache.setdefault(t, gen_free_name(self.tree, self.env, 'T'))
+        self.env[n] = t
+        return n
+
+
+def get_io_from_annotations(f, tree, defn_env):
+    inputs = []
+    outputs = []
+    tuple_out = False
+    namer = TypeNamer(tree, defn_env)
+    for k, v in f.__annotations__.items():
+        if k == 'return':
+            if isinstance(v, tuple):
+                tuple_out = True
+                for idx, t in enumerate(v):
+                    t_name = namer(t)
+                    outputs.append((f'O{idx}', t_name))
+            else:
+                t_name = namer(v)
+                outputs.append(('O', t_name))
+        else:
+            t_name = namer(v)
+            inputs.append((k, t_name))
+
+    return inputs, outputs, tuple_out
+
+
+def gen_io_args(inputs, outputs, async_reset, magma_name):
     io_args = []
     for name, type_ in inputs:
-        type_ = astor.to_source(type_).rstrip()
         io_args.append(f"{name}={magma_name}.In({type_})")
     io_args.append(f"CLK={magma_name}.In({magma_name}.Clock)")
     if async_reset:
         io_args.append(f"ASYNCRESET={magma_name}.In({magma_name}.AsyncReset)")
-    if isinstance(output_type, ast.Tuple):
-        outputs = []
-        for i, elem in enumerate(output_type.elts):
-            output_type_str = astor.to_source(elem).rstrip()
-            io_args.append(f"O{i}={magma_name}.Out({output_type_str})")
-    else:
-        output_type_str = astor.to_source(output_type).rstrip()
-        io_args.append(f"O={magma_name}.Out({output_type_str})")
+    for name, type_ in outputs:
+        io_args.append(f"{name}={magma_name}.Out({type_})")
     return ', '.join(io_args)
 
 
@@ -327,7 +362,7 @@ def gen_product_call_args(name, comb_out_count, eval_type, keys, comb_out_wiring
             arrays = "".join([f"['{k}']" for k in keys + [key]])
             dots = ".".join(keys + [key])
             comb_out_wiring.append(
-                f"{name}{arrays}.I <= comb_out[{comb_out_count}].{dots} \n"
+                f"{name}{arrays}.I @= comb_out[{comb_out_count}].{dots} \n"
             )
             call_args.append(f"{key}={name}{arrays}.O")
     return f"{magma_name}.namedtuple(" + ", ".join(call_args) + ")"
@@ -348,15 +383,14 @@ def _sequential(
     magma_name = gen_free_name(call_def, defn_env, 'm')
     defn_env[magma_name] = m
 
-    inputs, output_type = get_io(call_def)
-    io_args = gen_io_args(inputs, output_type, async_reset, magma_name)
+    inputs, outputs, tuple_out = get_io_from_annotations(cls.__call__, call_def, defn_env)
+    io_args = gen_io_args(inputs, outputs, async_reset, magma_name)
 
     circuit_combinational_output_type = []
     circuit_combinational_args = []
     circuit_combinational_call_args = []
     comb_out_wiring = []
     for name, type_ in inputs:
-        type_ = astor.to_source(type_).rstrip()
         circuit_combinational_args.append(f"{name}: {type_}")
         circuit_combinational_call_args.append(f"{name}=io.{name}")
 
@@ -373,37 +407,38 @@ def _sequential(
                 circuit_combinational_call_args.append(f"self_{name}_O=" + call_args)
             else:
                 circuit_combinational_call_args.append(f"self_{name}_O={name}")
-                comb_out_wiring.append(f"{name}.I <= comb_out[{comb_out_count}]\n")
+                comb_out_wiring.append(f"{name}.I @= comb_out[{comb_out_count}]\n")
             comb_out_count += 1
         else:
             for key, value in eval_value.interface.ports.items():
                 if isinstance(value, (m.Clock, m.AsyncReset)):
                     continue
-                type_ = repr(type(value))
+                t_name = gen_free_name(call_def, defn_env, 'T')
+                defn_env[t_name] = type(value)
                 if value.is_output():
-                    circuit_combinational_args.append(f"self_{name}_{value}: {magma_name}.{type_}")
-                    circuit_combinational_call_args.append(f"self_{name}_{value}={name}.{value}")
+                    circuit_combinational_args.append(f"self_{name}_{key}: {t_name}")
+                    circuit_combinational_call_args.append(f"self_{name}_{key}={name}.{key}")
                 if value.is_input():
-                    circuit_combinational_output_type.append(f"{magma_name}.{type_}")
-                    comb_out_wiring.append(f"{name}.{value} <= comb_out[{comb_out_count}]\n")
+                    circuit_combinational_output_type.append(t_name)
+                    comb_out_wiring.append(f"{name}.{key} @= comb_out[{comb_out_count}]\n")
                     comb_out_count += 1
 
     circuit_combinational_args = ', '.join(circuit_combinational_args)
     circuit_combinational_call_args = ', '.join(circuit_combinational_call_args)
 
-    if isinstance(output_type, ast.Tuple):
-        output_types = []
-        for i, elem in enumerate(output_type.elts):
-            circuit_combinational_output_type.append(astor.to_source(elem).rstrip())
-            comb_out_wiring.append(f"io.O{i} <= comb_out[{comb_out_count + i}]\n")
+    if tuple_out:
+        for i, (n, t) in enumerate(outputs):
+            circuit_combinational_output_type.append(t)
+            comb_out_wiring.append(f"io.{n} @= comb_out[{comb_out_count + i}]\n")
     else:
-        output_type_str = astor.to_source(output_type).rstrip()
-        circuit_combinational_output_type.append(output_type_str)
+        assert len(outputs) == 1
+        n, t = outputs[0]
+        circuit_combinational_output_type.append(t)
         # Handle case when no registers, so only one output
         index_str = ""
         if comb_out_count > 0:
             index_str = f"[{comb_out_count}]"
-        comb_out_wiring.append(f"io.O <= comb_out{index_str}\n")
+        comb_out_wiring.append(f"io.O @= comb_out{index_str}\n")
 
     tab = 4 * ' '
     comb_out_wiring = (2 * tab).join(comb_out_wiring)
