@@ -1,39 +1,19 @@
-import magma as m
-
-
-primitive_to_python_operator_name_map = {
-    "and": "and_",
-    "or": "or_",
-    "xor": "xor",
-    "shl": "lshift",
-    "lshr": "rshift",
-    "ashr": "rshift",
-    "urem": "mod",
-    "srem": "mod",
-    "udiv": "floordiv",
-    "sdiv": "floordiv",
-    "ule": "le",
-    "ult": "lt",
-    "uge": "ge",
-    "ugt": "gt",
-    "sle": "le",
-    "slt": "lt",
-    "sge": "ge",
-    "sgt": "gt",
-    "not": "invert"
-}
+from magma.array import Array
+from magma.bit import Bit
+from magma.bits import Bits
+from magma.tuple import Tuple
 
 
 def BitOrBits(width):
     if width is None:
-        return m.Bit
+        return Bit
     if not isinstance(width, int):
         raise ValueError(f"Expected width to be None or int, got {width}")
-    return m.Bits[width]
+    return Bits[width]
 
 
 def pretty_str(t):
-    if issubclass(t, m.Tuple):
+    if issubclass(t, Tuple):
         args = []
         for i in range(len(t)):
             key_str = str(list(t.keys())[i])
@@ -47,9 +27,9 @@ def pretty_str(t):
         # Insert first newline + indent and last newline
         result = "\n    " + result + "\n"
         s = f"Tuple({result})"
-    elif issubclass(t, m.Bits):
+    elif issubclass(t, Bits):
         s = str(t)
-    elif issubclass(t, m.Array):
+    elif issubclass(t, Array):
         s = f"Array[{t.N}, {pretty_str(t.T)}]"
     else:
         s = str(t)
